import collections
import json
from datetime import datetime
import time
import pytz
from typing import Dict
import ast

from requests import Session
import urllib.parse as _urlencode

from ksxt.base.types import IndexType
from ksxt.models.common import CommonResponse, CommonResponseHeader


class Exchange:
    id = None
    name = None
    version = None
    is_dev = False

    session = None
    # 세션 유효 시간 (초)
    session_lifetime = 10
    # 세션의 마지막 사용 시간
    session_last_used: time = None

    timeout = 10000  # milliseconds = seconds * 1000
    synchronous = True

    required_credentials = {
        "open_key": False,
        "secret_key": False,
        "uid": False,
        "login": False,
        "password": False,
        "token": False,
    }

    def __init__(self):
        if not self.session and self.synchronous:
            self.session = Session()

    def __del__(self):
        if self.session:
            try:
                self.session.close()
            except Exception as e:
                pass

    def __str__(self):
        return self.name

    # region public feeder
    def fetch_markets(self, market_name: str):
        """
        Market 정보 조회

        Args:
            market_name (str, optional): Market 구분 코드.
        """
        pass

    def fetch_security(self, symbol: str, base_market: str = "KRW"):
        """
        종목 정보 조회

        Args:
            symbol (str): 종목코드
            base_market (str, optional): Market 구분 코드. Defaults to 'KRW'.
        """
        pass

    def fetch_ticker(self, symbol: str, base_market: str = "KRW"):
        """
        시세 정보 조회

        Args:
            symbol (str): 종목코드
            base_market (str, optional): Market 구분 코드. Defaults to 'KRW'.
        """
        pass

    def fetch_historical_data(
        self,
        symbol: str,
        time_frame: str,
        start: datetime | None = None,
        end: datetime | None = None,
        base_market: str = "KRW",
    ):
        """
        과거 봉 정보 조회

        Args:
            symbol (str): 종목코드
            time_frame (str): 봉조회기준
            start (datetime | None, optional): 조회 시작일. Defaults to None.
            end (datetime | None, optional): 조회 종료일. Defaults to None.
            base_market (str, optional): Market 구분 코드. Defaults to 'KRW'.
        """
        pass

    def resample(self, df, timeframe: str, offset):
        ohlcv = {"open": "first", "high": "max", "low": "min", "close": "last", "volume": "sum"}

        result = df.resample(timeframe.upper(), offset=offset).apply(ohlcv)
        return result

    def fetch_is_holiday(self, dt: datetime, base_market: str = "KRW"):
        """
        휴장일 조회

        Args:
            dt (datetime): 조회 날짜 (YYYYMMDD)
            base_market (str, optional): Market 구분 코드. Defaults to 'KRW'.
        """
        pass

    # endregion public feeder

    # region private feeder
    def fetch_user_info(self, base_market: str = "KRW"):
        """
        회원 정보 조회

        Args:
            base_market (str, optional): Market 구분 코드. Defaults to 'KRW'.
        """
        pass

    def fetch_trade_fee(self, symbol: str, base_market: str = "KRW"):
        """
        종목의 거래 수수료 정보 조회

        Args:
            symbol (str): 종목코드
            base_market (str, optional): Market 구분 코드. Defaults to 'KRW'.
        """
        pass

    def fetch_balance(self, acc_num: str, base_market: str = "KRW", excluded_symbols: list[str] | None = None):
        """
        보유 자산 조회

        Args:
            acc_num (str): 계좌 번호
            base_market (str, optional): Market 구분 코드. Defaults to 'KRW'.
            excluded_symbols (list[str] | None, optional): 제외할 종목 리스트. Defaults to None.
        """
        pass

    def fetch_cash(self, acc_num: str, base_market: str = "KRW"):
        """
        예수금 조회

        Args:
            acc_num (str): 계좌 번호
            base_market (str, optional): Market 구분 코드. Defaults to 'KRW'.
        """
        pass

    def fetch_screener_list(self, base_market: str = "KRW"):
        """
        조건식 리스트 조회

        Returns:
            _type_: 조건식 리스트
        """
        pass

    def fetch_screener(self, screen_id: str, base_market: str = "KRW"):
        """
        조건식 조회 결과

        Args:
            screen_id (str): Screener 조회 값 (조건식 조회 결과)
            base_market (str, optional): Market 구분 코드. Defaults to 'KRW'.
        """
        pass

    def fetch_deposit_history(
        self, acc_num: str, start: datetime | None = None, end: datetime | None = None, base_market: str = "KRW"
    ):
        """
        입금 내역 조회

        Args:
            acc_num (str): 계좌 번호
            start (datetime | None, optional): 조회 시작일. Defaults to None.
            end (datetime | None, optional): 조회 종료일. Defaults to None.
            base_market (str, optional): Market 구분 코드. Defaults to 'KRW'.
        """
        pass

    def fetch_withdrawal_history(
        self, acc_num: str, start: datetime | None = None, end: datetime | None = None, base_market: str = "KRW"
    ):
        """
        출금 내역 조회

        Args:
            acc_num (str): 계좌 번호
            start (datetime | None, optional): 조회 시작일. Defaults to None.
            end (datetime | None, optional): 조회 종료일. Defaults to None.
            base_market (str, optional): Market 구분 코드. Defaults to 'KRW'.
        """
        pass

    # endregion private feeder

    # region broker
    def create_order(
        self,
        acc_num: str,
        symbol: str,
        ticket_type: str,
        price: float,
        qty: float,
        amount: float,
        otype: str,
        base_market: str = "KRW",
    ):
        """
        주문 발행

        Args:
            acc_num (str): 계좌정보(계좌번호, 지갑정보)
            symbol (str): 종목정보(종목코드)
            ticket_type (str): EntryLong, EntryShort, ExitLong, ExitShort, ...
            price (float): 가격 (지정가 시 필수)
            qty (float): 수량 (매도 시 필수.)
            amount (float): 총금액 (시장가 매수 시 필수)
            otype (str): market(시장가), limit(지정가), ...
            base_market (str, optional): Market 구분 코드. Defaults to 'KRW'.
        """
        pass

    def cancel_order(
        self, acc_num: str, order_id: str, symbol: str | None = "", qty: float = 0, *args, base_market: str = "KRW"
    ):
        """
        미체결 주문 취소

        Args:
            acc_num (str): 계좌정보(계좌번호, 지갑정보)
            order_id (str): 주문 정보(주문 id)
            symbol (str | None, optional): 종목정보(종목코드). Defaults to ''.
            qty (float, optional): 수량. Defaults to 0..
            base_market (str, optional): Market 구분 코드. Defaults to 'KRW'.
        """
        pass

    def fetch_open_order(
        self,
        acc_num: str,
        symbol: str | None = "",
        start: datetime | None = None,
        end: datetime | None = None,
        base_market: str = "KRW",
    ):
        """
        미체결 주문 내역 조회

        Args:
            acc_num (str): 계좌정보(계좌번호, 지갑정보)
            symbol (str | None, optional): 종목정보(종목코드) Defaults to ''.
            start (datetime | None, optional): 조회 시작일자. Defaults to None.
            end (datetime | None, optional): 조회 종료일자. Defaults to None.
            base_market (str, optional): Market 구분 코드. Defaults to 'KRW'.
        """
        pass

    def fetch_closed_order(
        self,
        acc_num: str,
        symbol: str | None = "",
        start: datetime | None = None,
        end: datetime | None = None,
        base_market: str = "KRW",
    ):
        """
        체결 주문 내역 조회

        Args:
            acc_num (str): 계좌정보(계좌번호, 지갑정보)
            symbol (str | None, optional): 종목정보(종목코드) Defaults to ''.
            start (datetime | None, optional): 조회 시작일자. Defaults to None.
            end (datetime | None, optional): 조회 종료일자. Defaults to None.
            base_market (str, optional): Market 구분 코드. Defaults to 'KRW'.
        """
        pass

    def reserve_order(
        self, acc_num: str, symbol: str, price: float, qty: float, target_date: str, base_market: str = "KRW"
    ):
        """
        예약 주문 발행

        Args:
            acc_num (str): 계좌정보(계좌번호, 지갑정보)
            symbol (str): 종목정보(종목코드)
            price (float): 가격
            qty (float): 수량
            target_date (str): 예약일자
            base_market (str, optional): Market 구분 코드. Defaults to 'KRW'.
        """
        pass

    # endregion broker

    def create_common_response(self, success: str, msg_code: str, msg: str, info: dict) -> CommonResponse:
        return CommonResponse(success=success, msg_code=msg_code, msg=msg, info=info)

    def create_common_header(self, request_params: dict) -> CommonResponseHeader:
        return CommonResponseHeader(request_params=request_params)

    # region utils
    @staticmethod
    def now(base_market: str = "KRW"):
        if base_market == "KRW":
            return datetime.now(tz=pytz.timezone("Asia/Seoul"))
        elif base_market == "USD":
            return datetime.now(tz=pytz.timezone("US/Eastern"))
        else:
            return datetime.now(tz=pytz.utc)

    @staticmethod
    def seconds():
        return int(time.time())

    @staticmethod
    def milliseconds():
        return int(time.time() * 1_000)

    @staticmethod
    def set_attr(self, attrs):
        for key in attrs:
            if hasattr(self, key) and isinstance(getattr(self, key), dict):
                setattr(self, key, Exchange.deep_extend(getattr(self, key), attrs[key]))
            else:
                setattr(self, key, attrs[key])

    @staticmethod
    def extend(*args):
        if args is not None:
            result = None
            if type(args[0]) is collections.OrderedDict:
                result = collections.OrderedDict()
            else:
                result = {}

            for arg in args:
                result.update(arg)

            return result

        return {}

    @staticmethod
    def deep_extend(*args):
        result = None
        for arg in args:
            if isinstance(arg, dict):
                if not isinstance(result, dict):
                    result = {}
                for key in arg:
                    result[key] = Exchange.deep_extend(result[key] if key in result else None, arg[key])
            else:
                result = arg

        return result

    @staticmethod
    def omit(d, *args):
        if isinstance(d, dict):
            result = d.copy()
            for arg in args:
                if type(arg) is list:
                    for key in arg:
                        if key in result:
                            del result[key]
                else:
                    if arg in result:
                        del result[arg]
            return result
        return d

    @staticmethod
    def implode_params(string, params):
        if isinstance(params, dict):
            for key in params:
                if not isinstance(params[key], list):
                    string = string.replace("{" + key + "}", str(params[key]))
        return string

    @staticmethod
    def urlencode(params={}, doseq=False):
        new_params = params.copy()
        for key, value in params.items():
            if isinstance(value, bool):
                new_params[key] = "true" if value else "false"
        return _urlencode.urlencode(new_params, doseq, quote_via=_urlencode.quote)

    def calculate_rate_limiter_cost(self, api, method, path, params, config={}):
        return self.safe_value(config, "cost", 1)

    def parse_json(self, http_response):
        return json.loads(http_response, parse_float=str, parse_int=str)

    # region safe method
    @staticmethod
    def key_exists(dictionary, key):
        if hasattr(dictionary, "__getitem__") and not isinstance(dictionary, str):
            if isinstance(dictionary, list) and type(key) is not int:
                return False
            try:
                value = dictionary[key]
                return value is not None and value != ""
            except LookupError:
                return False
        return False

    @staticmethod
    def safe_value(dictionary, key, default_value=None):
        return dictionary[key] if Exchange.key_exists(dictionary, key) else default_value

    @staticmethod
    def safe_string(dictionary, key, default_value=""):
        return str(dictionary[key]) if Exchange.key_exists(dictionary, key) else default_value

    @staticmethod
    def safe_number(dictionary, key, default_value=0):
        value = Exchange.safe_string(dictionary, key)
        if value == "":
            return default_value

        try:
            return float(value)
        except Exception:
            return default_value

    @staticmethod
    def safe_boolean(dictionary, key, default_value=False):
        value = Exchange.safe_string(dictionary, key)
        if value == "":
            return default_value

        try:
            return bool(ast.literal_eval(value.capitalize()))
        except (ValueError, SyntaxError):
            return default_value

    # endregion safe method

    @staticmethod
    def keysort(dictionary):
        return collections.OrderedDict(sorted(dictionary.items(), key=lambda t: t[0]))

    @staticmethod
    def sort_by(array, key, descending=False):
        return sorted(array, key=lambda k: k[key] if k[key] is not None else "", reverse=descending)

    @staticmethod
    def sort_by_2(array, key1, key2, descending=False):
        return sorted(
            array,
            key=lambda k: (k[key1] if k[key1] is not None else "", k[key2] if k[key2] is not None else ""),
            reverse=descending,
        )

    @staticmethod
    def index_by(array, key):
        result = {}
        if type(array) is dict:
            array = Exchange.keysort(array).values()
        is_int_key = isinstance(key, int)
        for element in array:
            if ((is_int_key and (key < len(element))) or (key in element)) and (element[key] is not None):
                k = element[key]
                result[k] = element
        return result

    @staticmethod
    def to_array(value):
        return list(value.values()) if type(value) is dict else value

    def filter_by_array(self, objects, key: IndexType, values=None, indexed=True):
        objects = self.to_array(objects)

        # return all of them if no values were passed
        if values is None or not values:
            return self.index_by(objects, key) if indexed else objects

        results = []
        for i in range(0, len(objects)):
            if self.in_array(objects[i][key], values):
                results.append(objects[i])

        return self.index_by(results, key) if indexed else results

    # endregion



from datetime import datetime, timezone
from operator import attrgetter
from typing import Dict, List, Optional

from ksxt.models.transaction import (
    ClosedOrderHistory,
    ClosedOrderInfo,
    DepositHistory,
    OpenedOrderHistory,
    OpenedOrderInfo,
    TransactionInfo,
    WithdrawalHistory,
)
from ksxt.models.balance import BalanceData, BalanceInfo
from ksxt.models.cash import CashInfo
from ksxt.models.historical import HistoricalDataInfo, OHLCVData
from ksxt.models.market import MarketInfo, SecurityData, TradeFeeInfo
from ksxt.models.order import CancelOrderResponse, CreateOrderResponse
from ksxt.models.orderbook import MultiSymbolOrderBookInfos, OrderBookData, OrderBookInfo
from ksxt.models.ticker import MultiSymbolTickerInfo, TickerInfo
from ksxt.parser.parser import BaseParser
from ksxt.utils import safer, sorter


class BithumbParser(BaseParser):
    def safe_symbol(self, base_market: str, security: str) -> str:
        return f"{base_market}-{security}"

    def parse_markets(self, response: List[Dict], base_market: str = "KRW") -> MarketInfo:
        securities = {}
        for market in response:
            market_base = market["market"].split("-")[0]
            if market_base != base_market:
                continue

            security_data = self._parse_market(market)
            securities[security_data.symbol] = security_data

        return MarketInfo(market_id=base_market, securities=securities)

    def _parse_market(self, market: Dict) -> SecurityData:
        return SecurityData(
            symbol=safer.safe_string(market, "market"),
            name=safer.safe_string(market, "korean_name"),
            type="crypto",
            warning_code=False if safer.safe_string(market, "market_warning") == "NONE" else True,
        )

    def parse_historical_data(
        self,
        response: List[Dict],
        symbol: str,
        start: datetime | None = None,
        end: datetime | None = None,
        base_market: str = "KRW",
    ) -> HistoricalDataInfo:
        ohlcv = [self._parse_ohlcva(_, base_market) for _ in response]

        # Filter by start and end datetime
        if start:
            ohlcv = [data for data in ohlcv if data.datetime >= start]
        if end:
            ohlcv = [data for data in ohlcv if data.datetime <= end]

        sorted_ohlcv = sorter.sort_by(ohlcv, key="datetime")

        security_name = symbol

        result = HistoricalDataInfo(
            symbol=self.safe_symbol(base_market, security_name), security_name=security_name, history=sorted_ohlcv
        )

        return result

    def _parse_ohlcva(self, ohlcva: dict, base_market: str = "KRW") -> OHLCVData:
        fmt = "%Y-%m-%dT%H:%M:%S"
        return OHLCVData(
            datetime=datetime.strptime(safer.safe_string(ohlcva, "candle_date_time_kst"), fmt),
            open_price=safer.safe_number(ohlcva, "opening_price"),
            high_price=safer.safe_number(ohlcva, "high_price"),
            low_price=safer.safe_number(ohlcva, "low_price"),
            close_price=safer.safe_number(ohlcva, "trade_price"),
            acml_volume=safer.safe_number(ohlcva, "candle_acc_trade_volume"),
            acml_amount=safer.safe_number(ohlcva, "candle_acc_trade_price"),
        )

    def parse_ticker(self, response: List[Dict], base_market: str = "KRW") -> TickerInfo:
        ticker_info = self._parse_ticker(response[0], base_market=base_market)
        return ticker_info

    def parse_tickers(self, response: List[Dict], base_market: str = "KRW") -> MultiSymbolTickerInfo:
        tickers_info = {}

        for ticker_data in response:
            symbol = safer.safe_string(ticker_data, "market")
            ticker_info = self._parse_ticker(ticker_data, base_market)
            tickers_info[symbol] = ticker_info

        return MultiSymbolTickerInfo(tickers=tickers_info)

    def _parse_ticker(self, response: dict, base_market: str) -> TickerInfo:
        timestamp_ms = safer.safe_number(response, "timestamp")
        timestamp_s = timestamp_ms / 1000.0

        return TickerInfo(
            symbol=safer.safe_string(response, "market"),
            price=safer.safe_number(response, "trade_price"),
            ts=datetime.fromtimestamp(timestamp_s),
        )

    def parse_orderbook(self, response: List[Dict], base_market: str = "KRW") -> OrderBookInfo:
        ask_list = []
        bid_list = []

        for item in response:
            for index, data in enumerate(item["orderbook_units"]):
                ask_data, bid_data = self._parse_orderbook(data, index, base_market)
                ask_list.append(ask_data)
                bid_list.append(bid_data)

        return OrderBookInfo(
            total_asks_qty=safer.safe_string(response[0], "total_ask_size"),
            total_bids_qty=safer.safe_string(response[0], "total_bid_size"),
            asks=ask_list,
            bids=bid_list,
        )

    def parse_orderbooks(self, response: dict, base_market: str) -> MultiSymbolOrderBookInfos:
        order_books = {}

        # response에서 'order_books' 항목이 리스트 형태로 되어 있다고 가정
        for symbol_info in response:
            symbol = symbol_info["market"]  # 각 종목의 심볼 추출
            orderbook_units = symbol_info["orderbook_units"]  # 호가 정보 추출

            ask_list = []
            bid_list = []

            for index, data in enumerate(orderbook_units):
                ask_data, bid_data = self._parse_orderbook(data, index, base_market)
                ask_list.append(ask_data)
                bid_list.append(bid_data)

            # OrderBookInfo 객체 생성
            order_book_info = OrderBookInfo(
                total_asks_qty=safer.safe_number(symbol_info, "total_ask_size"),
                total_bids_qty=safer.safe_number(symbol_info, "total_bid_size"),
                asks=ask_list,
                bids=bid_list,
            )

            # 심볼과 함께 OrderBookInfo를 딕셔너리에 추가
            order_books[symbol] = order_book_info

        # MultiSymbolOrderBookInfo 객체 반환
        return MultiSymbolOrderBookInfos(order_books=order_books)

    def _parse_orderbook(
        self, orderbook: dict, index: int, base_market: str = "KRW"
    ) -> tuple[OrderBookData, OrderBookData]:
        ask_data = OrderBookData(
            side="ask",
            ob_num=index,
            ob_price=safer.safe_number(orderbook, "ask_price"),
            ob_qty=safer.safe_number(orderbook, "ask_size"),
        )

        bid_data = OrderBookData(
            side="bid",
            ob_num=index,
            ob_price=safer.safe_number(orderbook, "bid_price"),
            ob_qty=safer.safe_number(orderbook, "bid_size"),
        )

        return ask_data, bid_data

    def parse_balance(
        self, response: List[Dict], base_market: str = "KRW", excluded_symbols: Optional[list[str]] = None
    ) -> BalanceInfo:
        # 1. Filter out entries where 'currency' is not the base_market
        # 2. Ensure 'unit_currency' is the base_market
        # 3. Exclude entries where 'currency' is in excluded_symbols
        filtered_data = [
            data
            for data in response
            if data["currency"] != base_market
            and data["unit_currency"] == base_market
            and (excluded_symbols is None or data["currency"] not in excluded_symbols)
        ]

        # Initialize balance list
        balances = [self._parse_balance(data, base_market) for data in filtered_data]

        # 총 매입금액
        total_amount = sum(bal.price * bal.qty for bal in balances)
        total_evaluation_amount = sum(bal.evaluation_price * bal.qty for bal in balances)
        total_pnl_amount = total_evaluation_amount - total_amount
        total_pnl_ratio = (total_pnl_amount / total_amount * 100) if total_amount != 0 else 0

        return BalanceInfo(
            currency=base_market,
            total_amount=total_amount,
            total_evaluation_amount=total_amount,
            total_pnl_amount=total_pnl_amount,
            total_pnl_ratio=total_pnl_ratio,
            balances=balances,
        )

    def _parse_balance(self, data: Dict, base_market: str = "KRW") -> BalanceData:
        return BalanceData(
            symbol=self.safe_symbol(base_market, safer.safe_string(data, "currency")),
            name=self.safe_symbol(base_market, safer.safe_string(data, "currency")),
            evaluation_price=0,  # 필요에 따라 적절히 설정
            price=safer.safe_number(data, "avg_buy_price"),
            pnl_amount=0,  # 필요에 따라 적절히 설정
            pnl_ratio=0,  # 필요에 따라 적절히 설정
            qty=safer.safe_number(data, "balance"),
            free_qty=safer.safe_number(data, "balance") - safer.safe_number(data, "locked"),
            used_qty=safer.safe_number(data, "locked"),
        )

    def parse_cash(self, response: List[Dict], base_market: str = "KRW") -> CashInfo:
        # 1. Filter entries where 'currency' is equal to 'base_market'
        # 2. Ensure 'unit_currency' is also equal to 'base_market'
        filtered_data = [
            data for data in response if data["currency"] == base_market and data["unit_currency"] == base_market
        ]

        # 필터링된 데이터가 없으면 기본값으로 CashInfo 반환
        if not filtered_data:
            return CashInfo(currency=base_market, cash=0.0, cash_d1=0.0, cash_d2=0.0)

        # 필터링된 첫 번째 항목을 가져와서 CashInfo 생성
        data = filtered_data[0]

        return CashInfo(
            currency=base_market,
            cash=safer.safe_number(data, "balance"),
            cash_d1=safer.safe_number(data, "balance"),  # 데이터를 가정하여 같은 값 사용
            cash_d2=safer.safe_number(data, "balance"),  # 데이터를 가정하여 같은 값 사용
        )

    def parse_security(self, response: dict, base_market: str = "KRW") -> SecurityData:
        market_info = response.get("market", {})

        return SecurityData(
            symbol=safer.safe_string(market_info, "id"),
            name=safer.safe_string(market_info, "name"),
            type="crypto",
            bid_qty_unit=safer.safe_number(market_info["bid"], "price_unit"),
            ask_qty_unit=safer.safe_number(market_info["ask"], "price_unit"),
            bid_min_qty=0,
            ask_min_qty=0,
            bid_max_qty=0,
            ask_max_qty=0,
            bid_amount_unit=1,
            ask_amount_unit=1,
            bid_min_amount=safer.safe_number(market_info["bid"], "min_total"),
            ask_min_amount=safer.safe_number(market_info["ask"], "min_total"),
            bid_max_amount=safer.safe_number(market_info, "max_total"),
            ask_max_amount=safer.safe_number(market_info, "max_total"),
            warning_code=safer.safe_string(market_info, "state") != "active",
        )

    def parse_trade_fee(self, response: dict, base_market: str = "KRW") -> TradeFeeInfo:
        return TradeFeeInfo(
            limit_bid_fee=safer.safe_number(response, "bid_fee"),
            limit_ask_fee=safer.safe_number(response, "ask_fee"),
            market_bid_fee=safer.safe_number(response, "maker_bid_fee"),
            market_ask_fee=safer.safe_number(response, "maker_ask_fee"),
        )

    def parse_open_order_history(
        self, response: List[Dict], start: datetime | None = None, end: datetime | None = None, base_market: str = "KRW"
    ) -> OpenedOrderHistory:
        orders = [self._parse_open_order_info(order, base_market) for order in response]

        # Filter by start and end datetime
        if start:
            orders = [
                order for order in orders if order.created_at.astimezone(timezone.utc) >= start.astimezone(timezone.utc)
            ]
        if end:
            orders = [
                order for order in orders if order.created_at.astimezone(timezone.utc) <= end.astimezone(timezone.utc)
            ]

        return OpenedOrderHistory(history=orders)

    def _parse_open_order_info(self, order: dict, base_market: str = "KRW") -> OpenedOrderInfo:
        return OpenedOrderInfo(
            uuid=safer.safe_string(order, "uuid"),
            type=safer.safe_string(order, "side"),
            symbol=self.safe_symbol(base_market, safer.safe_string(order, "market")),
            price=safer.safe_number(order, "price"),
            qty=safer.safe_number(order, "volume"),
            amount=safer.safe_number(order, "price") * safer.safe_number(order, "volume"),
            tax=0,
            fee=safer.safe_number(order, "reserved_fee"),
            created_at=datetime.fromisoformat(safer.safe_string(order, "created_at")),
            order_type=safer.safe_string(order, "ord_type"),
        )

    def parse_closed_order_history(
        self, response: List[Dict], start: datetime | None = None, end: datetime | None = None, base_market: str = "KRW"
    ) -> ClosedOrderHistory:
        orders = [self._parse_closed_order_info(order, base_market) for order in response]

        # Filter by start and end datetime
        if start:
            orders = [
                order for order in orders if order.created_at.astimezone(timezone.utc) >= start.astimezone(timezone.utc)
            ]
        if end:
            orders = [
                order for order in orders if order.created_at.astimezone(timezone.utc) <= end.astimezone(timezone.utc)
            ]

        return ClosedOrderHistory(history=orders)

    def _parse_closed_order_info(self, order: dict, base_market: str = "KRW") -> ClosedOrderInfo:
        return ClosedOrderInfo(
            uuid=safer.safe_string(order, "uuid"),
            type=safer.safe_string(order, "side"),
            symbol=safer.safe_string(order, "market"),
            price=safer.safe_number(order, "price"),
            qty=safer.safe_number(order, "volume"),
            amount=safer.safe_number(order, "price") * safer.safe_number(order, "volume"),
            tax=0,
            fee=safer.safe_number(order, "reserved_fee"),
            created_at=datetime.fromisoformat(safer.safe_string(order, "created_at")),
            order_type=safer.safe_string(order, "ord_type"),
        )

    def parse_cancel_order(self, response: dict, base_market: str = "KRW") -> CancelOrderResponse:
        return CancelOrderResponse(order_datetime=datetime.now(), order_id=safer.safe_string(response, "uuid"))

    def parse_create_order(self, response: dict, base_market: str = "KRW") -> CreateOrderResponse:
        return CreateOrderResponse(order_datetime=datetime.now(), order_id=safer.safe_string(response, "uuid"))

    def parse_withdrawal_history(
        self, response: List[Dict], start: datetime | None = None, end: datetime | None = None, base_market: str = "KRW"
    ) -> WithdrawalHistory:
        parsed_items = [self._parse_transaction_history_item(item, base_market) for item in response]

        # Filter by start and end datetime
        if start:
            parsed_items = [
                item
                for item in parsed_items
                if item.created_at.astimezone(timezone.utc) >= start.astimezone(timezone.utc)
            ]
        if end:
            parsed_items = [
                item
                for item in parsed_items
                if item.created_at.astimezone(timezone.utc) <= end.astimezone(timezone.utc)
            ]

        return WithdrawalHistory(history=parsed_items)

    def parse_deposit_history(
        self, response: List[Dict], start: datetime | None = None, end: datetime | None = None, base_market: str = "KRW"
    ) -> DepositHistory:
        parsed_items = [self._parse_transaction_history_item(item, base_market) for item in response]

        # Filter by start and end datetime
        if start:
            parsed_items = [
                item
                for item in parsed_items
                if item.created_at.astimezone(timezone.utc) >= start.astimezone(timezone.utc)
            ]
        if end:
            parsed_items = [
                item
                for item in parsed_items
                if item.created_at.astimezone(timezone.utc) <= end.astimezone(timezone.utc)
            ]

        return DepositHistory(history=parsed_items)

    def _parse_transaction_history_item(self, item: Dict, base_market: str = "KRW") -> TransactionInfo:
        return TransactionInfo(
            uuid=safer.safe_string(item, "uuid"),
            type=safer.safe_string(item, "type"),
            amount=safer.safe_number(item, "amount"),
            tax=0,
            fee=safer.safe_number(item, "fee"),
            currency=safer.safe_string(item, "currency"),
            created_at=datetime.fromisoformat(safer.safe_string(item, "created_at")),
        )



from datetime import datetime
from typing import Dict, List, Optional

from ksxt.models.balance import BalanceData, BalanceInfo
from ksxt.models.cash import CashInfo
from ksxt.models.historical import HistoricalDataInfo, OHLCVData
from ksxt.models.market import HolidayHistory, HolidayInfo, MarketInfo, SecurityData, TradeFeeInfo
from ksxt.models.order import CancelOrderResponse, CreateOrderResponse, ModifyOrderResponse
from ksxt.models.orderbook import OrderBookData, OrderBookInfo
from ksxt.models.ticker import MultiSymbolTickerInfo, TickerInfo
from ksxt.models.token import TokenInfo
from ksxt.models.transaction import (
    ClosedOrderHistory,
    ClosedOrderInfo,
    DepositHistory,
    OpenedOrderHistory,
    OpenedOrderInfo,
)
from ksxt.parser.parser import BaseParser
from ksxt.utils import safer, sorter, timer


class KoreaInvestParser(BaseParser):
    def safe_symbol(self, base_market: str, symbol: str) -> str:
        return symbol

    def parse_token(self, response: dict) -> TokenInfo:
        date_string = safer.safe_string(response, "access_token_token_expired")
        date_format = "%Y-%m-%d %H:%M:%S"
        expired_date = datetime.strptime(date_string, date_format)

        return TokenInfo(
            access_token=safer.safe_string(response, "access_token"),
            token_type=safer.safe_string(response, "token_type"),
            remain_time_second=safer.safe_number(response, "expires_in"),
            expired_datetime=expired_date,
        )

    def parse_markets(self, response: dict, base_market: str = "KRW") -> MarketInfo:
        securities = {}
        for market in response:
            security_data = self._parse_market(market, base_market)
            securities[security_data.symbol] = security_data

        return MarketInfo(market_id=base_market, securities=securities)

    def _parse_market(self, market: Dict, base_market: str = "KRW") -> SecurityData:
        return SecurityData(
            symbol=safer.safe_string(market, "market"),
            name=safer.safe_string(market, "korean_name"),
            type="stock",
            warning_code=safer.safe_boolean(market, " market_warning"),
        )

    def parse_historical_data(
        self,
        response: List[Dict],
        symbol: str,
        start: datetime | None = None,
        end: datetime | None = None,
        base_market: str = "KRW",
    ) -> HistoricalDataInfo:
        safe_response = safer.safe_value(response, "output2")
        ohlcv = [self._parse_ohlcva(_, base_market) for _ in safe_response]

        # Filter by start and end datetime
        if start:
            ohlcv = [data for data in ohlcv if data.datetime >= start]
        if end:
            ohlcv = [data for data in ohlcv if data.datetime <= end]

        sorted_ohlcv = sorter.sort_by(ohlcv, key="datetime")

        security_name = symbol

        result = HistoricalDataInfo(
            symbol=self.safe_symbol(base_market, security_name), security_name=security_name, history=sorted_ohlcv
        )

        return result

    def _parse_ohlcva(self, ohlcva: Dict, base_market: str = "KRW") -> OHLCVData:
        fmt = "%Y%m%d"
        return OHLCVData(
            datetime=datetime.strptime(safer.safe_string(ohlcva, "stck_bsop_date"), fmt),
            open_price=safer.safe_number(ohlcva, "stck_oprc"),
            high_price=safer.safe_number(ohlcva, "stck_hgpr"),
            low_price=safer.safe_number(ohlcva, "stck_lwpr"),
            close_price=safer.safe_number(ohlcva, "stck_clpr"),
            acml_volume=safer.safe_number(ohlcva, "acml_vol"),
            acml_amount=safer.safe_number(ohlcva, "acml_tr_pbmn"),
        )

    def parse_historical_index_data(
        self,
        response: Dict,
        symbol: str,
        start: datetime | None = None,
        end: datetime | None = None,
        base_market: str = "KRW",
    ) -> HistoricalDataInfo:
        safe_response = safer.safe_value(response, "output2")
        ohlcv = [self._parse_index_ohlcva(_, base_market) for _ in safe_response]

        # Filter by start and end datetime
        if start:
            ohlcv = [data for data in ohlcv if data.datetime >= start]
        if end:
            ohlcv = [data for data in ohlcv if data.datetime <= end]

        sorted_ohlcv = sorter.sort_by(ohlcv, key="datetime")

        security_name = symbol

        result = HistoricalDataInfo(
            symbol=self.safe_symbol(base_market, security_name), security_name=security_name, history=sorted_ohlcv
        )

        return result

    def _parse_index_ohlcva(self, ohlcva: Dict, base_market: str = "KRW") -> OHLCVData:
        fmt = "%Y%m%d"
        return OHLCVData(
            datetime=datetime.strptime(safer.safe_string(ohlcva, "stck_bsop_date"), fmt),
            open_price=safer.safe_number(ohlcva, "bstp_nmix_oprc"),
            high_price=safer.safe_number(ohlcva, "bstp_nmix_hgpr"),
            low_price=safer.safe_number(ohlcva, "bstp_nmix_lwpr"),
            close_price=safer.safe_number(ohlcva, "bstp_nmix_prpr"),
            acml_volume=safer.safe_number(ohlcva, "acml_vol"),
            acml_amount=safer.safe_number(ohlcva, "acml_tr_pbmn"),
        )

    def parse_ticker(self, response: List[Dict], base_market: str = "KRW") -> TickerInfo:
        safe_response = safer.safe_value(response, "output")
        ticker_info = self._parse_ticker(safe_response, base_market)
        return ticker_info

    def _parse_ticker(self, response: dict, base_market: str) -> TickerInfo:
        return TickerInfo(
            symbol=safer.safe_string(response, "stck_shrn_iscd"),
            price=safer.safe_number(response, "stck_prpr"),
            ts=datetime.now(),
        )

    def parse_orderbook(self, response: List[Dict], base_market: str = "KRW") -> OrderBookInfo:
        ask_list = []
        bid_list = []

        level = 10

        safe_response = safer.safe_value(response, "output1")

        for index in range(1, level + 1):
            ask_data, bid_data = self._parse_orderbook(safe_response, index, base_market)
            ask_list.append(ask_data)
            bid_list.append(bid_data)

        return OrderBookInfo(
            total_asks_qty=safer.safe_number(safe_response, "total_askp_rsqn"),
            total_bids_qty=safer.safe_number(safe_response, "total_askp_rsqn"),
            asks=ask_list,
            bids=bid_list,
        )

    def _parse_orderbook(
        self, orderbook: Dict, index: int, base_market: str = "KRW"
    ) -> tuple[OrderBookData, OrderBookData]:
        ask_data = OrderBookData(
            side="ask",
            ob_num=index,
            ob_price=safer.safe_number(orderbook, f"askp{index}"),
            ob_qty=safer.safe_number(orderbook, f"askp_rsqn{index}"),
        )

        bid_data = OrderBookData(
            side="bid",
            ob_num=index,
            ob_price=safer.safe_number(orderbook, f"bidp{index}"),
            ob_qty=safer.safe_number(orderbook, f"bidp_rsqn{index}"),
        )

        return ask_data, bid_data

    def parse_balance(
        self, response: List[Dict], base_market: str = "KRW", excluded_symbols: Optional[list[str]] = None
    ) -> BalanceInfo:
        safe_response = safer.safe_value(response, "output1")

        # Initialize balance list with filtering
        balances = [
            self._parse_balance(_, base_market)
            for _ in safe_response
            if excluded_symbols is None or safer.safe_string(_, "pdno") not in excluded_symbols
        ]

        # 총 매입금액
        total_amount = sum(bal.price * bal.qty for bal in balances)
        total_evaluation_amount = sum(bal.evaluation_price * bal.qty for bal in balances)
        total_pnl_amount = total_evaluation_amount - total_amount
        total_pnl_ratio = (total_pnl_amount / total_amount * 100) if total_amount != 0 else 0

        return BalanceInfo(
            currency=base_market,
            total_amount=total_amount,
            total_evaluation_amount=total_evaluation_amount,
            total_pnl_amount=total_pnl_amount,
            total_pnl_ratio=total_pnl_ratio,
            balances=balances,
        )

    def _parse_balance(self, data: Dict, base_market: str = "KRW") -> BalanceData:
        return BalanceData(
            symbol=safer.safe_string(data, "pdno"),
            name=safer.safe_string(data, "prdt_name"),
            evaluation_price=safer.safe_number(data, "evlu_amt"),
            price=safer.safe_number(data, "pchs_avg_pric"),
            pnl_amount=safer.safe_number(data, "evlu_pfls_amt"),
            pnl_ratio=safer.safe_number(data, "evlu_pfls_rt"),
            qty=safer.safe_number(data, "hldg_qty"),
            free_qty=safer.safe_number(data, "ord_psbl_qty"),
            used_qty=safer.safe_number(data, "hldg_qty") - safer.safe_number(data, "ord_psbl_qty"),
        )

    def parse_cash(self, response: List[Dict], base_market: str = "KRW") -> CashInfo:
        safe_response = safer.safe_value(response, "output2")

        return CashInfo(
            currency=base_market,
            cash=safer.safe_number(safe_response, "dnca_tot_amt"),
            cash_d1=safer.safe_number(safe_response, "nxdy_excc_amt"),
            cash_d2=safer.safe_number(safe_response, "prvs_rcdl_excc_amt"),
        )

    def parse_security(self, response: Dict, base_market: str = "KRW") -> SecurityData:
        safe_response = safer.safe_value(response, "output")

        return SecurityData(
            symbol=safer.safe_string(safe_response, "pdno")[-6:],
            name=safer.safe_string(safe_response, "prdt_name"),
            type=safer.safe_string(safe_response, "mket_id_cd"),
            bid_qty_unit=0,  # TODO : where is information?
            ask_qty_unit=0,  # TODO : where is information?
            bid_min_qty=0,  # TODO : where is information?
            ask_min_qty=0,  # TODO : where is information?
            bid_max_qty=0,  # TODO : where is information?
            ask_max_qty=0,  # TODO : where is information?
            bid_amount_unit=1,  # TODO: where is information?
            ask_amount_unit=1,  # TODO : where is information?
            bid_min_amount=0,  # TODO : where is information?
            ask_min_amount=0,  # TODO : where is information?
            bid_max_amount=0,  # TODO : where is information?
            ask_max_amount=0,  # TODO : where is information?
            # 거래정지여부 or 관리종목여부
            warning_code=safer.safe_string(safe_response, "tr_stop_yn") == "Y"
            or safer.safe_string(safe_response, "admn_item_yn") == "Y",
        )

    def parse_trade_fee(self, response: Dict, base_market: str = "KRW") -> TradeFeeInfo:
        # API에서 조회되는 내역을 찾지 못해 홈페이지 안내 문을 작성함. (단위 : %)
        # https://securities.koreainvestment.com/main/customer/guide/_static/TF04ae010000.jsp

        # 거래수수료
        trading_fee = 0.18

        # 매매수수료
        return TradeFeeInfo(
            limit_bid_fee=0.0140527 + trading_fee,
            limit_ask_fee=0.0140527 + trading_fee,
            market_bid_fee=0.0140527 + trading_fee,
            market_ask_fee=0.0140527 + trading_fee,
        )

    def parse_closed_order_history(
        self, response: List[Dict], start: datetime | None = None, end: datetime | None = None, base_market: str = "KRW"
    ) -> ClosedOrderHistory:
        safe_response = safer.safe_value(response, "output1")
        orders = [self._parse_closed_order_info(_, base_market) for _ in safe_response]

        # Filter by start and end datetime
        if start:
            orders = [order for order in orders if order.created_at >= start]
        if end:
            orders = [order for order in orders if order.created_at <= end]

        return ClosedOrderHistory(history=orders)

    def _parse_closed_order_info(self, order: Dict, base_market: str = "KRW") -> ClosedOrderInfo:
        if safer.safe_string(order, "sll_buy_dvsn_cd") == "01":
            _type = "ask"
        else:
            _type = "bid"

        return ClosedOrderInfo(
            uuid=safer.safe_string(order, "odno"),
            type=_type,
            symbol=safer.safe_string(order, "pdno"),
            price=safer.safe_number(order, "avg_prvs"),
            qty=safer.safe_number(order, "tot_ccld_qty"),
            amount=safer.safe_number(order, "tot_ccld_amt"),
            tax=0,
            # 단일 주문 이력에는 수수료 관련 정보가 없고, 조회한 전체 목록에 대해 수수료 정보만 존재한다.
            fee=0,
            created_at=datetime.fromisoformat(safer.safe_string(order, "ord_dt")),
            order_type=safer.safe_string(order, "avg_prvs"),
        )

    def parse_open_order_history(
        self, response: List[Dict], start: datetime | None = None, end: datetime | None = None, base_market: str = "KRW"
    ) -> OpenedOrderHistory:
        safe_response = safer.safe_value(response, "output1")
        orders = [self._parse_open_order_info(_, base_market) for _ in safe_response]

        # Filter by start and end datetime
        if start:
            orders = [order for order in orders if order.created_at >= start]
        if end:
            orders = [order for order in orders if order.created_at <= end]

        return OpenedOrderHistory(history=orders)

    def _parse_open_order_info(self, order: Dict, base_market: str = "KRW") -> OpenedOrderInfo:
        if safer.safe_string(order, "sll_buy_dvsn_cd") == "01":
            _type = "ask"
        else:
            _type = "bid"

        return OpenedOrderInfo(
            uuid=safer.safe_string(order, "odno"),
            type=_type,
            symbol=safer.safe_string(order, "pdno"),
            price=safer.safe_number(order, "avg_prvs"),
            qty=safer.safe_number(order, "tot_ccld_qty"),
            amount=safer.safe_number(order, "tot_ccld_amt"),
            tax=0,
            # 단일 주문 이력에는 수수료 관련 정보가 없고, 조회한 전체 목록에 대해 수수료 정보만 존재한다.
            fee=0,
            created_at=datetime.fromisoformat(safer.safe_string(order, "ord_dt")),
            order_type=safer.safe_string(order, "avg_prvs"),
        )

    def parse_create_order(self, response: Dict, base_market: str = "KRW") -> CreateOrderResponse:
        safe_response = safer.safe_value(response, "output")
        fmt = "%Y%m%d"

        order_hhmmss = datetime.strptime(safer.safe_string(safe_response, "ORD_TMD"), fmt)
        order_datetime = timer.create_datetime_with_today(order_hhmmss)

        return CreateOrderResponse(order_datetime=order_datetime, order_id=safer.safe_string(safe_response, "ODNO"))

    def parse_cancel_order(self, response: Dict, base_market: str = "KRW") -> CancelOrderResponse:
        safe_response = safer.safe_value(response, "output")
        fmt = "%Y%m%d"

        order_hhmmss = datetime.strptime(safer.safe_string(safe_response, "ORD_TMD"), fmt)
        order_datetime = timer.create_datetime_with_today(order_hhmmss)

        return CancelOrderResponse(order_datetime=order_datetime, order_id=safer.safe_string(safe_response, "ODNO"))

    def parse_modify_order(self, response: Dict, base_market: str = "KRW") -> ModifyOrderResponse:
        safe_response = safer.safe_value(response, "output")
        fmt = "%Y%m%d"

        order_hhmmss = datetime.strptime(safer.safe_string(safe_response, "ORD_TMD"), fmt)
        order_datetime = timer.create_datetime_with_today(order_hhmmss)

        return ModifyOrderResponse(order_datetime=order_datetime, order_id=safer.safe_string(safe_response, "ODNO"))

    def parse_is_holiday(self, response: List[Dict], base_market: str = "KRW") -> HolidayHistory:
        safe_response = safer.safe_value(response, "output")
        history = [self._parse_is_holiday(_, base_market) for _ in safe_response]

        return HolidayHistory(market=["kospi", "kosdaq"], country="KR", history=history)

    def _parse_is_holiday(self, item: Dict, base_market: str = "KRW") -> HolidayInfo:
        return HolidayInfo(
            date=datetime.strptime(safer.safe_string(item, "bass_dt"), "%Y%m%d"),
            weekend=safer.safe_string(item, "wday_dvsn_cd"),
            is_holiday=not safer.safe_boolean(item, "opnd_yn"),
        )



from datetime import datetime
import time

import pytz


@staticmethod
def now(base_market: str = "KRW"):
    if base_market == "KRW":
        return datetime.now(tz=pytz.timezone("Asia/Seoul"))
    elif base_market == "USD":
        return datetime.now(tz=pytz.timezone("US/Eastern"))
    else:
        return datetime.now(tz=pytz.utc)


@staticmethod
def seconds():
    return int(time.time())


@staticmethod
def milliseconds():
    return int(time.time() * 1_000)


@staticmethod
def create_datetime_with_today(hhmmss: str) -> datetime:
    """
    hhmmss 형식의 시간 문자열을 받아 오늘 날짜를 포함한 datetime 객체를 생성합니다.

    Parameters:
    hhmmss (str): "hhmmss" 형식의 시간 문자열 (예: "123045")

    Returns:
    datetime: 오늘 날짜와 주어진 시간이 포함된 datetime 객체
    """
    # 현재 날짜를 가져옴
    today = datetime.now()

    # hhmmss 문자열을 datetime 객체로 변환
    time_part = datetime.strptime(hhmmss, "%H%M%S").time()

    # 현재 날짜와 주어진 시간을 합쳐 datetime 객체 생성
    combined_datetime = datetime.combine(today, time_part)

    return combined_datetime



from pydantic import BaseModel
from typing import List, Optional
from datetime import datetime

from ksxt.models.common import GeneralResponse


class OHLCVData(BaseModel):
    datetime: datetime
    open_price: float | int
    high_price: float | int
    low_price: float | int
    close_price: float | int
    acml_volume: float | int
    acml_amount: float | int


class HistoricalDataInfo(BaseModel):
    symbol: str
    security_name: Optional[str] = None
    history: List[OHLCVData]


# Historical Data에 대한 구체적인 Response 타입
class KsxtHistoricalDataResponse(GeneralResponse[HistoricalDataInfo]):
    pass



from datetime import datetime
from typing import Dict, List, Optional
from pydantic import BaseModel

from ksxt.models.common import GeneralResponse


class TradeFeeInfo(BaseModel):
    limit_bid_fee: float
    limit_ask_fee: float

    market_bid_fee: float
    market_ask_fee: float


class SecurityData(BaseModel):
    # 종목코드
    symbol: str
    # 종목이름
    name: str
    # 종목유형
    type: str

    # 매수 주문 수량 단위
    bid_qty_unit: Optional[float] = 0
    # 매도 주문 수량 단위
    ask_qty_unit: Optional[float] = 0

    # 최소 매수 주문 수량
    bid_min_qty: Optional[float] = 0
    # 최소 매도 주문 수량
    ask_min_qty: Optional[float] = 0
    # 최대 매수 주문 수량
    bid_max_qty: Optional[float] = 0
    # 최대 매도 주문 수량
    ask_max_qty: Optional[float] = 0

    # 매수 주문 금액 단위
    bid_amount_unit: Optional[float] = 0
    # 매도 주문 금액 단위
    ask_amount_unit: Optional[float] = 0

    # 최소 매수 주문 금액
    bid_min_amount: Optional[float] = 0
    # 최소 매도 주문 금액
    ask_min_amount: Optional[float] = 0
    # 최대 매수 주문 금액
    bid_max_amount: Optional[float] = 0
    # 최대 매도 주문 금액
    ask_max_amount: Optional[float] = 0

    warning_code: bool = False


class MarketInfo(BaseModel):
    market_id: str
    securities: Dict[str, SecurityData]


class HolidayInfo(BaseModel):
    date: datetime
    # 01 : 일요일 ~ 07: 토요일
    weekend: str
    is_holiday: bool


class HolidayHistory(BaseModel):
    market: List[str]
    country: str
    history: List[HolidayInfo]


class KsxtTradeFeeResponse(GeneralResponse[TradeFeeInfo]):
    pass


class KsxtSecurityResponse(GeneralResponse[SecurityData]):
    pass


class KsxtMarketResponse(GeneralResponse[MarketInfo]):
    pass


class KsxtHolidayResponse(GeneralResponse[HolidayHistory]):
    pass



from .transaction import (
    KsxtWithdrawalHistoryResponse,
    KsxtDepositHistoryResponse,
    KsxtOpenOrderResponse,
    KsxtClosedOrderResponse,
)
from .balance import KsxtBalanceResponse
from .cash import KsxtCashResponse
from .error import KsxtErrorResponse
from .historical import KsxtHistoricalDataResponse
from .market import KsxtTradeFeeResponse, KsxtSecurityResponse, KsxtMarketResponse, KsxtHolidayResponse
from .order import KsxtCreateOrderResponse, KsxtCancelOrderResponse, KsxtModifyOrderResponse
from .orderbook import KsxtSingleOrderBookResponse, KsxtMultiOrderBookResponse
from .ticker import KsxtTickerResponse, KsxtTickersResponse

from .token import KsxtTokenResponse



from ksxt.base.exchange import Exchange


class ComExchange(Exchange):
    def __init__(self, id: str, name: str, zone: str = "Asia/Seoul", is_dev: bool = False) -> None:
        super().__init__(id=id, name=name, zone=zone, is_dev=is_dev)

        self.is_session_alive = False



from datetime import datetime
from typing import Generic, Iterable, Iterator, Literal, TypeVar, get_args
from ksxt.market.base import MarketItemBase, MarketBase
from ksxt.market.db import Base

STIS = {"1": "Index", "2": "Stock", "3": "ETP(ETF)", "4": "Warrant"}

SJONG = {"0": "구성종목없음", "1": "구성종목있음"}

ETYP = {
    "001": "ETF",
    "002": "ETN",
    "003": "ETC",
    "004": "Others",
    "005": "VIX Underlying ETF",
    "006": "VIX Underlying ETN",
}


class UsStockItem(MarketItemBase):
    # Notional code
    ncod: str
    # Exchange ID
    exid: str
    # Exchange Code
    excd: str
    # Exchange Name
    exnm: str
    # Symbol
    symb: str
    # Realtime Symbol
    rsym: str
    # Korean name
    knam: str
    # English name
    enam: str
    # Security Type
    stis: Literal["1", "2", "3", "4"]
    # currency
    curr: str
    # float position
    zdiv: str
    # data type
    ztyp: str
    # base price
    base: float
    # bid order size
    bnit: float
    # ask ordersize
    anit: float
    # market start time (HHMM)
    mstm: str
    # market end time (HHMM)
    metm: str
    # DR 여부 : Y/N
    isdr: bool
    # DR 국가코드
    drcd: str
    # 업종분류코드
    icod: str
    # 지수구성종목 존재 여부
    sjong: Literal["0", "1"]
    # Tick Size Type
    ttyp: str
    #
    etyp: Literal["001", "002", "003", "004", "005", "006"]
    # Tick size type 상세 (ttyp 8일 경우 사용) : 런던 제트라 유로넥스트
    ttyp_sb: str

    @property
    def stis_name(self) -> str:
        return STIS.get(self.stis, None)

    @property
    def sjong_name(self) -> str:
        return SJONG.get(self.sjong, None)

    @property
    def etyp_name(self) -> str:
        return ETYP.get(self.etyp, None)

    def __init__(self, data: str):
        super().__init__(data, delimiter="\t")


t_item = TypeVar("t_item", bound=UsStockItem)
t_dbitem = TypeVar("t_dbitem", bound=Base)


class UsStockMarket(Generic[t_item, t_dbitem], MarketBase[t_item, t_dbitem]):
    def search(self, name: str, limit: int = 50) -> Iterable[t_item]:
        """
        종목 검색 (이름 기준)

        Args:
            name (str): 종목의 영어 이름
            limit (int, optional): 조회 갯수 제한. Defaults to 50.

        Returns:
            Iterable[t_item]: 종목 정보
        """
        _, db_type = get_args(self.__orig_bases__[0])  # type: ignore
        return self._search(db_type.enam, name, limit)  # type: ignore

    def items(self, offset: int = 0, limit: int = 100) -> Iterable[t_item]:
        """
        종목 조회 (offset 기준)

        Args:
            offset (int, optional): offset 정보. Defaults to 0.
            limit (int, optional): 조회 갯수 제한. Defaults to 100.

        Returns:
            Iterable[t_item]: 종목 정보
        """
        return super().items(offset, limit)

    def all(self) -> Iterator[t_item]:
        """
        모든 종목 조회

        Returns:
            _type_: 종목 정보

        Yields:
            Iterator[t_item]: 종목 정보
        """
        return super().all()  # type: ignore

    def __getitem__(self, code: str) -> t_item:
        """
        종목 검색 (종목코드 기준)

        Args:
            code (str): 종목 코드

        Returns:
            t_item: 종목 정보
        """
        """종목을 가져옵니다."""
        _, db_type = get_args(self.__orig_bases__[0])  # type: ignore
        return self._get(db_type.symb, code)



from datetime import date, datetime, timedelta
import logging
import os.path as path
import tempfile
from threading import Lock, Thread
import time
from typing import Iterable, Optional, Union
from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker, Session
from itertools import chain

from ksxt.market.db import Base, Market
from ksxt.market.base import MarketBase
from ksxt.market.krx.kosdaq import Kosdaq
from ksxt.market.krx.kospi import Kospi
from ksxt.market.krx.stock import KrxStockItem
from ksxt.market.logging import KsxtLogger
from ksxt.market.markets import MARKET_ITEM_TYPE, MARKET_TYPE, MARKETS
from ksxt.market.us.amex import Amex
from ksxt.market.us.nasdaq import Nasdaq
from ksxt.market.us.nyse import Nyse
from ksxt.market.us.stock import UsStockItem


class MarketManager(KsxtLogger):
    db_path: str
    engine: Engine
    markets: dict[str, MarketBase]

    _sessionmaker: sessionmaker
    _rth: int = 0
    _sstd: datetime
    _sl: Lock

    @property
    def kospi(self) -> Kospi:
        """코스피"""
        return self["kospi"]  # type: ignore

    @property
    def kosdaq(self) -> Kosdaq:
        """코스닥"""
        return self["kosdaq"]  # type: ignore

    @property
    def nasdaq(self) -> Nasdaq:
        """나스닥(National Association of Securities Dealers Automated Quotations)"""
        return self["nasdaq"]  # type: ignore

    @property
    def nyse(self) -> Nyse:
        """뉴욕증권거래소(the New York Stock Exchange)"""
        return self["nyse"]  # type: ignore

    @property
    def amex(self) -> Amex:
        """미국증권거래소(AMerican stock EXchange)"""
        return self["amex"]  # type: ignore

    @property
    def session(self) -> Session:
        """세션"""
        return self._sessionmaker()

    def stock(self, code: str) -> Union[KrxStockItem, UsStockItem]:
        """코스피/코스닥/뉴욕/나스닥/아멕스 종목"""
        return self.kospi[code] or self.kosdaq[code] or self.nyse[code] or self.nasdaq[code] or self.amex[code]  # type: ignore

    def __init__(
        self,
        db_path: Optional[str] = None,
        auto_sync: bool = False,
    ):
        if db_path is None:
            db_path = path.join(tempfile.gettempdir(), f".ksxt-cache_market.{datetime.now().strftime('%Y%m%d')}.db")

        self.db_path = db_path
        self.markets = {}
        self._sstd = None
        self._sl = Lock()
        self._create_database()

        if auto_sync:
            self._st_sync()

    def _create_database(self):
        self.engine = create_engine(
            f"sqlite:///{self.db_path}?check_same_thread=False"
        )  # maybe it will be ok for a while...
        Base.metadata.create_all(self.engine)
        self._sessionmaker = sessionmaker(bind=self.engine, autoflush=False)

    def _logger_ready(self, logger: logging.Logger):
        logger.debug(f"database open: {self.db_path}")

    def _init(self):
        self._sync_all(comp=False, verbose=False)

    def sync_at(self, code: str) -> datetime:
        """종목 정보를 마지막으로 업데이트한 시간"""
        return self.session.query(Market.sync_at).filter(Market.code == code).scalar()

    def sync_all(self):
        """모든 시장을 동기화합니다."""
        self._sync_all(comp=True, verbose=False)

    def _sync_all(self, comp: bool = False, verbose: bool = False) -> bool:
        with self._sl:
            now = datetime.now()

            if comp:
                self._sstd = now
            else:
                if self._sstd:
                    if now - self._sstd < timedelta(seconds=30):
                        if verbose:
                            self.logger.info("MARKET: too early to sync")

                        return False

                self._sstd = now

        for code in MARKETS:
            self[code].sync()

        return True

    def _get_market(self, session: Session, code: str) -> Market:
        market = self[code]

        if not market:
            raise KeyError(f"Unknown market: {code}")

        session.merge(market._market())
        session.commit()

        return session.query(Market).filter(Market.code == code).one()

    def _update_sync_at(self, code: str):
        """종목 정보를 업데이트한 시간을 업데이트합니다."""
        with self.session as sess:
            dm = self._get_market(sess, code)
            dm.sync_at = datetime.now()  # type: ignore
            sess.commit()

    def _st_sync(self):
        """자동동기화 스레드"""
        self._rth += 1
        tid = self._rth
        Thread(target=self._at_sync, args=(tid,), daemon=True).start()

    def _at_sync(self, tid: int):
        """자동 동기화"""
        n = False
        time.sleep(31)
        while self._rth == tid:
            try:
                if self._sync_all(comp=False, verbose=n):
                    self.logger.info("MARKET: auto-sync")
            except Exception as e:
                self.logger.error(f"MARKET: auto-sync exception: {e}")
            finally:
                if not n:
                    n = True
                time.sleep((60 * 60 * 24) + 5)

    def __getitem__(self, code: str) -> MARKET_TYPE:
        """시장"""
        if code not in self.markets:
            market = MARKETS[code](self)
            market._emit_logger(self.logger)
            self.markets[code] = market
        else:
            market = self.markets[code]
        return market  # type: ignore

    def search(self, keyword: str, origin: list[str] = None) -> dict[str, Iterable[MARKET_ITEM_TYPE]]:
        """종목 검색을 수행합니다.

        Args:
            keyword: 검색 키워드
            origin: 검색할 시장 (기본값: 모든 시장)
        """
        result = {}

        for market in self.markets.values():
            if not origin or market.code in origin:
                result[market.code] = market.search(keyword)

        return result

    def stock_search(self, keyword: str) -> dict[str, Iterable[KrxStockItem]]:
        """코스피/코스닥 종목 검색을 수행합니다."""
        return self.search(keyword, ["kospi", "kosdaq"])  # type: ignore

    def stock_search_combined(self, keyword: str) -> Iterable[KrxStockItem]:
        """코스피/코스닥 종목 검색을 수행합니다."""
        return chain(*self.stock_search(keyword).values())

    def search_one(self, keyword: str, origin: list[str] = None) -> MARKET_ITEM_TYPE:
        """종목 검색을 수행합니다.

        Args:
            keyword: 검색 키워드
            origin: 검색할 시장 (기본값: 모든 시장)
        """
        for market in self.markets.values():
            if not origin or market.code in origin:
                result = market.search_one(keyword)

                if result:
                    return result

        return None

    def stock_search_one(self, keyword: str) -> KrxStockItem:
        return self.search_one(keyword, ["kospi", "kosdaq"])  # type: ignore



import asyncio
import time
import aiohttp
from datetime import datetime
from typing import Any, Dict, Literal, Optional


from ksxt.base.errors import NotSupportedError
from ksxt.base.rate_limiter import RateLimiterContext
from ksxt.base.rest_exchange import RestExchange
from ksxt.config import VALID_METHODS
import ksxt.models


class AsyncExchange(RestExchange):
    synchronous = False

    def __init__(self, config: Dict = None, filename: str = None):
        super().__init__(config, filename)

        self.asyncio_loop = None
        self.session: aiohttp.ClientSession = None

    async def initialize(self):
        if self.asyncio_loop is None:
            self.asyncio_loop = asyncio.get_event_loop()

        if self.session is None or (
            self.session_last_used and (time.time() - self.session_last_used > self.session_lifetime)
        ):
            if self.session:
                await self.session.close()
            self.session = aiohttp.ClientSession()
            self.session_last_used = time.time()

    async def close(self):
        if self.session:
            await self.session.close()
            self.session = None
            self.session_last_used = None

    async def __aenter__(self):
        await self.initialize()
        return self

    async def __aexit__(self, *args):
        await self.close()

    async def fetch(self, url, method="GET", headers=None, body=None, params=None):
        # Ensure that resources are initialized before the request
        await self.initialize()

        method_lower = method.lower()
        if method_lower not in VALID_METHODS:
            raise ValueError(f"Invalid HTTP method: {method}")

        session_method = getattr(self.session, method.lower())

        # TODO : Set rate limiters value when config load
        api_name = ""

        if api_name and api_name in self.rate_limiters:
            await self.rate_limiters[api_name].async_acquire()

        try:
            async with session_method(
                url,
                headers=headers,
                data=str(body).encode() if body else None,
                params=params,
                timeout=aiohttp.ClientTimeout(total=int(self.timeout / 1000)),
            ) as response:
                http_response = await response.text(errors="replace")
                json_response = self.parse_json(http_response)
                return json_response
        except (asyncio.TimeoutError, aiohttp.ClientError) as e:
            details = f"{self.id} {method} {url}"
            raise TimeoutError(details) from e
        finally:
            if (
                api_name
                and api_name in self.rate_limiters
                and isinstance(self.rate_limiters[api_name], RateLimiterContext)
            ):
                self.rate_limiters[api_name].release()

    async def fetch2(
        self, path, security_type, params={}, headers: Any | None = None, body: Any | None = None, config={}
    ):
        is_activate = self.apis[self.type][security_type][path]["activate"]
        if not is_activate:
            return {
                "response": {
                    # 성공 실패 여부
                    "success": "-1",
                    # 응답코드
                    "code": "fail",
                    # 응답메세지
                    "message": f"지원하지 않는 함수({path}) 입니다.",
                }
            }

        # if self.enableRateLimit:
        #     cost = self.calculate_rate_limiter_cost(api, method, path, params, config)
        #     self.throttle(cost)

        # self.lastRestRequestTimestamp = self.milliseconds()

        method_type = self.apis[self.type][security_type][path]["method"]
        api_type = self.apis[self.type][security_type][path]["api"]
        request = self.sign(path, security_type, method_type, api_type, headers, body, params, config)
        return await self.fetch(
            request["url"], request["method"], request["headers"], request["body"], request["params"]
        )

    # region base method
    async def fetch_markets(self, market_name: str) -> ksxt.models.KsxtMarketResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_markets.__qualname__}() is not supported yet.")

    async def fetch_security(self, symbol: str, base_market: str = "KRW") -> ksxt.models.KsxtSecurityResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_security.__qualname__}() is not supported yet.")

    async def fetch_ticker(self, symbol: str, base_market: str = "KRW") -> ksxt.models.KsxtTickerResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_ticker.__qualname__}() is not supported yet.")

    async def fetch_orderbook(self, symbol: str, base_market: str = "KRW") -> ksxt.models.KsxtSingleOrderBookResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_orderbook.__qualname__}() is not supported yet.")

    async def fetch_historical_data(
        self,
        symbol: str,
        time_frame: str,
        start: datetime | None = None,
        end: datetime | None = None,
        base_market: str = "KRW",
    ) -> ksxt.models.KsxtHistoricalDataResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_historical_data.__qualname__}() is not supported yet.")

    async def fetch_is_holiday(self, dt: datetime, base_market: str = "KRW"):
        raise NotSupportedError(f"{self.id} {self.fetch_is_holiday.__qualname__}() is not supported yet.")

    async def fetch_user_info(self, base_market: str = "KRW"):
        raise NotSupportedError(f"{self.id} {self.fetch_user_info.__qualname__}() is not supported yet.")

    async def fetch_balance(
        self, acc_num: str, base_market: str = "KRW", excluded_symbols: list[str] | None = None
    ) -> ksxt.models.KsxtBalanceResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_balance.__qualname__}() is not supported yet.")

    async def fetch_cash(self, acc_num: str, base_market: str = "KRW") -> ksxt.models.KsxtCashResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_cash.__qualname__}() is not supported yet.")

    async def fetch_screener_list(self, base_market: str = "KRW"):
        raise NotSupportedError(f"{self.id} {self.fetch_screener_list.__qualname__}() is not supported yet.")

    async def fetch_screener(self, screen_id: str, base_market: str = "KRW"):
        raise NotSupportedError(f"{self.id} {self.fetch_screener.__qualname__}() is not supported yet.")

    async def fetch_deposit_history(
        self, acc_num: str, start: datetime | None = None, end: datetime | None = None, base_market: str = "KRW"
    ) -> ksxt.models.KsxtDepositHistoryResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_deposit_history.__qualname__}() is not supported yet.")

    async def fetch_withdrawal_history(
        self, acc_num: str, start: datetime | None = None, end: datetime | None = None, base_market: str = "KRW"
    ) -> ksxt.models.KsxtWithdrawalHistoryResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_withdrawal_history.__qualname__}() is not supported yet.")

    async def create_order(
        self,
        acc_num: str,
        symbol: str,
        ticket_type: Literal["EntryLong", "EntryShort", "ExitLong", "ExitShort"],
        otype: Literal["limit", "market"],
        price: float | None = 0,
        qty: float | None = 0,
        amount: float | None = 0,
        base_market: str = "KRW",
    ) -> ksxt.models.KsxtCreateOrderResponse:
        raise NotSupportedError(f"{self.id} {self.create_order.__qualname__}() is not supported yet.")

    async def cancel_order(
        self, acc_num: str, order_id: str, symbol: str | None = "", qty: float = 0, *args, base_market: str = "KRW"
    ) -> ksxt.models.KsxtCancelOrderResponse:
        raise NotSupportedError(f"{self.id} {self.cancel_order.__qualname__}() is not supported yet.")

    async def modify_order(
        self,
        acc_num: str,
        order_id: str,
        price: float,
        qty: float,
        *args,
        symbol: str | None = "",
        base_market: str = "KRW",
    ):
        raise NotSupportedError(f"{self.id} {self.modify_order.__qualname__}() is not supported yet.")

    async def fetch_open_order(
        self,
        acc_num: str,
        symbol: str | None = "",
        start: datetime | None = None,
        end: datetime | None = None,
        base_market: str = "KRW",
    ) -> ksxt.models.KsxtOpenOrderResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_open_order.__qualname__}() is not supported yet.")

    async def fetch_closed_order(
        self,
        acc_num: str,
        symbol: str | None = "",
        start: datetime | None = None,
        end: datetime | None = None,
        base_market: str = "KRW",
    ) -> ksxt.models.KsxtClosedOrderResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_closed_order.__qualname__}() is not supported yet.")

    async def reserve_order(
        self, acc_num: str, symbol: str, price: float, qty: float, target_date: str, base_market: str = "KRW"
    ):
        raise NotSupportedError(f"{self.id} {self.reserve_order.__qualname__}() is not supported yet.")

    # endregion base method

    async def sleep(self, milliseconds):
        return await asyncio.sleep(milliseconds / 1000)



from datetime import datetime
from pydantic import BaseModel
from ksxt.models.common import GeneralResponse


class TokenInfo(BaseModel):
    access_token: str
    token_type: str
    remain_time_second: float
    expired_datetime: datetime


class KsxtTokenResponse(GeneralResponse[TokenInfo]):
    pass
